/*
 * Copyright (c) 2015-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jun  2 2015
 */
#include <string.h>

#include "libgenomix.h"


/* --- genomix_server ------------------------------------------------------ */

void __attribute__ ((visibility ("default")))
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  char emsg[256];
  char key[128];
  int s;

  if (nrhs < 1)
    mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

  mxGetString(prhs[0], key, sizeof(key));
  if (!strcmp(key, "get")) {
    char uri[1024];
    const char *r;
    int id;

    if (nrhs < 3)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

    id = mxGetScalar(prhs[1]);
    mxGetString(prhs[2], uri, sizeof(uri));

    if (nrhs > 3) {
      if (!mxIsStruct(prhs[3]))
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
      {
        int n = mxGetNumberOfFields(prhs[3]);
        const char *args[2*n+1];
        int argsfree[n];
        mxArray *v;
        int i, err;

        err = 0;
        for(i = 0; i < n; i++) {
          args[2*i] = mxGetFieldNameByNumber(prhs[3], i);
          v = mxGetFieldByNumber(prhs[3], 0, i);

          args[2*i+1] = NULL;
          argsfree[i] = 0;
          if (err) continue;

          if (mxIsStruct(v)) {
            /* convert struct to json */
            char *json = s2j_input(v, emsg, sizeof(emsg));
            if (!json) { err = 1; continue; }

            args[2*i+1] = json;
            argsfree[i] = 1;

          } else if (mxIsChar(v)) {
            /* assume properly encoded data */
            args[2*i+1] = mxArrayToString(v);
            if (!args[2*i+1]) { err = 1; continue; }
            argsfree[i] = 2;
          }
        }
        args[2*n] = NULL;

        if (!err) err = genomix_get(id, uri, args, emsg, sizeof(emsg));
        r = err ? NULL : genomix_reply(id, NULL, emsg, sizeof(emsg));

        for(i = 0; i < n; i++) {
          if (!args[2*i+1]) continue;
          switch (argsfree[i]) {
            case 2: mxFree((char *)args[2*i+1]); break;
            case 1: free((char *)args[2*i+1]); break;
          }
        }
      }
    } else {
      if (genomix_get(id, uri, NULL, emsg, sizeof(emsg)))
        r = NULL;
      else
        r = genomix_reply(id, NULL, emsg, sizeof(emsg));
    }

    if (!r)
      mexErrMsgIdAndTxt("genomix:run", emsg);

    if (plhs > 0) {
      if (r[0] == '{' || r[0] == '[' || r[0] == '"')
        plhs[0] = j2s_value(&r, emsg, sizeof(emsg));
      else
        plhs[0] = mxCreateString(r);
      if (!plhs[0]) mexErrMsgIdAndTxt("genomix:run", emsg);
    }

  } else if (!strcmp(key, "doevents")) {
    int timeout = 0;
    int id;

    if (nrhs < 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    if (nrhs > 2) {
      double *t;
      if (!mxIsDouble(prhs[2]))
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
      t = mxGetPr(prhs[2]);
      if (t) timeout = *t * 1000;
    }
    if (genomix_doevent(id, timeout))
      mexErrMsgIdAndTxt("genomix:run", "disconnected client");

  } else if (!strcmp(key, "json")) {
    const char *in, *json;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    in = mxArrayToString(prhs[1]);
    if (!in)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

    json = in;
    if (nlhs > 0) {
      plhs[0] = j2s_value(&json, emsg, sizeof(emsg));
      if (!plhs[0]) mexErrMsgIdAndTxt("genomix:run", emsg);
    }
    mxFree((void *)in);


  } else if (!strcmp(key, "connect")) {
    char *host = NULL;
    char *session = NULL;

    if (nrhs > 3)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    if (nrhs > 2) {
      session = mxArrayToString(prhs[2]);
      if (!session)
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    }
    if (nrhs > 1) {
      host = mxArrayToString(prhs[1]);
      if (!host)
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    }

    s = genomix_connect(host, session, emsg, sizeof(emsg));

    if (host) mxFree(host);
    if (session) mxFree(session);
    if (s < 0) mexErrMsgIdAndTxt("genomix:run", emsg);

    if (!mexIsLocked()) {
      mexLock();
      if (genomix_debug) printf("INFO: locking " PACKAGE_STRING "\n");
    }
    if (nlhs > 0)
      plhs[0] = mxCreateDoubleScalar(s);

  } else if (!strcmp(key, "connected")) {
    int id;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    if (nlhs > 0)
      plhs[0] = mxCreateDoubleScalar(genomix_connected(id));

  } else if (!strcmp(key, "disconnect")) {
    int id;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    if (!genomix_free(id)) {
      mexUnlock();
      if (genomix_debug) printf("INFO: unlocking " PACKAGE_STRING "\n");
    }

  } else
    mexErrMsgIdAndTxt("genomix:args", "invalid command");
}


/* --- genomix_event ------------------------------------------------------- */

void
genomix_event(const char *data)
{
  mxArray *prhs[1];
  char emsg[256];

  prhs[0] = j2s_value(&data, emsg, sizeof(emsg));
  mexCallMATLAB(0, NULL, 1, prhs, "genomix.client.doevents");
}
